unit RtToken;

{ RTToken                                                       }
{ Copyright 1996 R&T Software, Tomasz Stanczak                  }
{ CIS 100735,3273                                               }
{ WWW http://ourworld.compuserve.com/homepages/tomasz           }
{                                                               }
{ This unit contains a TRTTokenList class, a general use class  }
{ for analyzing and parsing any text information into tokens.   }
{ You can define delimiters and whitespaces, and then take any  }
{ text apart. Personally, I have never needed anything more     }
{ advanced or complex. You can parse SQL expressions or fields  }
{ during importing text files into the database.                }
{                                                               }
{ constuctor Create(Value: PChar);                              }
{   creates a RTTokenList from a PChar pointer. The property    }
{   DontFree decides, if the memory should be freed by the List.}
{                                                               }
{ constuctor CreateFromString(constValue: string);              }
{   creates a RTTokenList from a Pascal string. It copies its   }
{   contents into internal buffer.                              }
{                                                               }
{ constuctor CreateFromStrings(Value: TStringList);             }
{   creates a RTTokenList from a PChar pointer retrieved using  }
{   the GetText method of the TStringList. The pointer returned }
{   is freed on destroy.                                        }
{                                                               }
{ procedure SetFromString(const Value: string);                 }
{   Resets the internal data sturctures and begins parsing of   }
{   Pascal string. It can be used during parsing of text file   }
{   read line by line without the need to create an object every}
{   time.                                                       }
{                                                               }
{ procedure SetFromStrings(Value: TStringList);                 }
{   Resets the internal data sturctures and begins parsing of   }
{   text buffer retrieved from the parameter.                   }
{                                                               }
{ function First: string;                                       }
{   Goes to the beggining and returns first token.              }
{                                                               }
{ function Next: string;                                        }
{   Retrieves next token.                                       }
{                                                               }
{ function Prior: string;                                       }
{   Retrieves prior token.                                      }
{                                                               }
{ function Last: string;                                        }
{   Retrieves last token, so all token until the end must be    }
{   parsed.                                                     }
{                                                               }
{ procedure Reset;                                              }
{   Resets internal data structures and disposes already parsed }
{   tokens.                                                     }
{                                                               }
{ property DontFree: Boolean;                                   }
{   decides if the memory pointed to by internal variables      }
{   belongs                                                     }
{   to the RTTokenList and should be freed.                     }
{                                                               }
{ property Index: longint;                                      }
{   current position within the token list.                     }
{                                                               }
{ property Count: word;                                         }
{   total number of tokens.                                     }
{                                                               }
{ property Token[Index: integer]: string;                       }
{   allows random acces to the token list. All token parsed are }
{   placed in the internal list, so that the next acces doesn't }
{   need a new parsing.                                         }
{                                                               }
{ property Ptr[Index: integer]: PChar;                          }
{   returnes a pointer to a specified token within the original }
{   text.                                                       }
{                                                               }
{ property Pos[Index: integer]: longint;                        }
{   retrieves the distance between the beginning of the text and}
{   specified token.                                            }
{                                                               }
{ property Delimiters: string;                                  }
{   defines token delimiters. Every delimiter starts new token. }
{                                                               }
{ property WhiteSpaces: string;                                 }
{   defines token whitespaces. Any number of whitespaces will be}
{   skipped without starting new token. A delimiter character   }
{   can                                                         }
{   be also a whitespace.                                       }
{                                                               }
{ property CurrentToken: string;                                }
{   returnes the last token being accessed.                     }
{                                                               }
{ property CurrentPos: longint;                                 }
{   distance between current position and the beginning of the  }
{   text.                                                       }
{                                                               }
{ property CurrentPtr: PChar;                                   }
{   pointer to the current position.                            }
{                                                               }
{ property Eof:Boolean;                                         }
{   are we at the end already?                                  }
{                                                               }
{ property OriginalPtr: PChar;                                  }
{   returns the pointer to the beggining of the text.           }
{                                                               }
{ This unit is a freeware, so you're granted an unlimited       }
{ license to use it for anything you want, except that you're   }
{ not allowed to distribute and/or market it as your own work,  }
{ alone or as a part of any component or component libraries in }
{ source code or compiled DCUs.                                 }
{                                                               }
{ All rights not explicitly granted here are reserved by the    }
{ author.                                                       }
{                                                               }
{ There are no warranties implied or explicite. You cannot make }
{ me responsible for anything you do with this unit and I won't }
{ promise to support it.                                        }

{------Fixed compiler switches----------------------------------------}
{$B-   Short-circuit boolean expressions }
{$G+   80286+ type instructions }
{$I+   I/O checking via exceptions }
{$V-   Disable var string checking }
{$W-   No Windows realmode stack frame }
{$X+   Enable extended syntax }
{---------------------------------------------------------------------}

interface

uses
   SysUtils, Classes;

type
	TRtTokenList = class(TObject)
   private
   	FToken: TList;
      FTokenPtr: TList;
   	FOriginalPtr: PChar;
      FPtr: PChar;
      FEof: Boolean;
		FDelim: array [0..63] of Char;
		FWhite: array [0..63] of Char;
	   FDLen: longint;
	   FTLen: longint;
      FCount: longint;
      FDontFree: Boolean;
      FCurrentToken: string;
      FIndex: longint;
      procedure SetPtr(Value: PChar);
      function GetCount: word;
      function GetToken(Index: integer): string;
      function GetPtr(Index: integer): PChar;
      function GetPos(Index: integer): longint;
      function GetDelimiters: string;
      procedure SetDelimiters(Value: string);
      function GetWhiteSpaces: string;
      procedure SetWhiteSpaces(Value: string);
      function GetCurrentPos: longint;
      function GetIndex: longint;
      procedure SetIndex(Value: longint);
   public
   	constructor Create(Value: PChar);
   	constructor CreateFromString(const Value: string);
   	constructor CreateFromStrings(Value: TStringList);
      destructor Destroy; override;
      procedure SetFromString(const Value: string);
      procedure SetFromStrings(Value: TStringList);
      function First: string;
      function Next: string;
      function Prior: string;
      function Last: string;
      procedure Reset;
      property DontFree: Boolean read FDontFree write FDontFree;
      property Index: longint read GetIndex write SetIndex;
      property Count: word read GetCount;
      property Token[Index: integer]: string read GetToken; default;
      property Ptr[Index: integer]: PChar read GetPtr;
      property Pos[Index: integer]: longint read GetPos;
      property Delimiters: string read GetDelimiters write SetDelimiters;
      property WhiteSpaces: string read GetWhiteSpaces write SetWhiteSpaces;
      property CurrentToken: string read FCurrentToken;
      property CurrentPos: longint read GetCurrentPos;
      property CurrentPtr: PChar read FPtr;
      property Eof:Boolean read FEof;
      property OriginalPtr: PChar read FOriginalPtr write SetPtr;
   end;

implementation

{ TRtTokenList }
constructor TRtTokenList.Create(Value: PChar);
begin
   FToken := TList.Create;
   FTokenPtr := TList.Create;
   StrPCopy(FDelim,' +-*=,;:/()[]{}#&%'+#0+#9+#10+#13+#34);
   FDontFree := True;
  	SetPtr(Value);
end;

constructor TRtTokenList.CreateFromString(const Value: string);
begin
	Create(nil);
   FDontFree := False;
	SetPtr(StrAlloc(Length(Value)+1));
   StrPCopy(FOriginalPtr,Value);
end;

constructor TRtTokenList.CreateFromStrings(Value: TStringList);
begin
	Create(nil);
   FDontFree := False;
	SetPtr(Value.GetText);
end;

destructor TRtTokenList.Destroy;
var
	i: integer;
begin
	if not FDontFree then
   	StrDispose(FOriginalPtr);
   for i:=FToken.Count-1 downto 0 do
   	if FToken[i]<>nil then
	   	DisposeStr(PString(FToken[i]));
   FToken.Free;
   FTokenPtr.Free;
end;

procedure TRtTokenList.SetPtr(Value: PChar);
begin
	FOriginalPtr := Value;
   Reset;
end;

function TRtTokenList.GetDelimiters: string;
begin
	Result := StrPas(FDelim);
end;

procedure TRtTokenList.SetDelimiters(Value: string);
begin
	StrPCopy(FDelim,Value);
   Reset;
end;

function TRtTokenList.GetWhiteSpaces: string;
begin
	Result := StrPas(FWhite);
end;

procedure TRtTokenList.SetWhiteSpaces(Value: string);
begin
	StrPCopy(FWhite,Value);
   Reset;
end;

function TRtTokenList.GetPtr(Index: integer): PChar;
begin
	if Index<FTokenPtr.Count then
	   Result := FTokenPtr[Index]
   else
	begin
      if FToken.Count=0 then
	   	First;
      while (FToken.Count<=Index) and (FCurrentToken<>'') do
      	Next;
      Result := FPtr;
   end;
end;

function TRtTokenList.GetPos(Index: integer): longint;
var
	Ptr: PChar;
begin
	if Index<FTokenPtr.Count then
	   Ptr := FTokenPtr[Index]
   else
	begin
      if FToken.Count=0 then
	   	First;
      while (FToken.Count<=Index) and (FCurrentToken<>'') do
      	Next;
      Ptr := FPtr;
   end;
	Result := Ptr-FOriginalPtr;
end;

function TRtTokenList.GetCurrentPos: longint;
begin
	Result := FPtr-FOriginalPtr;
end;

function TRtTokenList.GetIndex: longint;
begin
	Result := FIndex;
end;

procedure TRtTokenList.SetIndex(Value: longint);
begin
	if (Value<>FIndex) then
   begin
		if Value<=FToken.Count then
         FIndex := Value
      else
      begin
      	GetToken(Value);
         FIndex := FToken.Count;
      end;
   end;
end;

function TRtTokenList.GetCount: word;
var
   Ptr: PChar;
begin
   if FCount=0 then
   begin
	   Ptr := FOriginalPtr;
		while (Ptr^<>#0) do
      begin
   		while (Ptr^<>#0) and (StrScan(FWhite,Ptr^)<>nil) and (StrScan(FDelim,Ptr^)=nil) do
            Inc(Ptr);
   		if (Ptr^<>#0) and (StrScan(FWhite,Ptr^)=nil) then
         begin
     			Inc(FCount);
            {Inc(Ptr);}
     			while (Ptr^<>#0) and (StrScan(FDelim,Ptr^)=nil) do
	            Inc(Ptr);
	      end;
   		if (Ptr^<>#0) and (StrScan(FDelim,Ptr^)<>nil) then
  	      	Inc(Ptr);
		end;
   end;
   Result := FCount;
end;

function TRtTokenList.GetToken(Index: integer): string;
begin
	if Index<FToken.Count then
	   FCurrentToken := PString(FToken[Index])^
   else
	begin
      if FToken.Count=0 then
	   	First;
      while (FToken.Count<=Index) and (FCurrentToken<>'') do
      	Next;
   end;
   Result := FCurrentToken;
end;

procedure TRtTokenList.SetFromString(const Value: string);
begin
	if not FDontFree then
   	StrDispose(FOriginalPtr);
   FDontFree := False;
	SetPtr(StrAlloc(Length(Value)+1));
   StrPCopy(FOriginalPtr,Value);
end;

procedure TRtTokenList.SetFromStrings(Value: TStringList);
begin
	if not FDontFree then
   	StrDispose(FOriginalPtr);
   FDontFree := False;
	SetPtr(Value.GetText);
end;

function TRtTokenList.First: string;
begin
	Reset;
   Result := Next;
end;

function TRtTokenList.Next: string;
var
   Buf: PChar;
   Found: Boolean;
begin
	Found := False;
	Inc(FIndex);
  	FCurrentToken := '';
	if FIndex<FToken.Count then
	   FCurrentToken := PString(FToken[Index])^
   else while (not FEof) and (not Found) do
   begin
  		while (FPtr^<>#0) and (StrScan(FWhite,FPtr^)<>nil) and (StrScan(FDelim,FPtr^)=nil) do
			Inc(FPtr);
      Buf := FPtr;
  		if (FPtr^<>#0) and (StrScan(FWhite,FPtr^)=nil) then
      begin
  			while (FPtr^<>#0) and (StrScan(FDelim,FPtr^)=nil) do
            Inc(FPtr);
         {$IFDEF WIN32}
         SetLength(FCurrentToken,FPtr-Buf);
	      StrLCopy(@FCurrentToken[1],Buf,FPtr-Buf);
         {$ELSE}                              
	      StrLCopy(@FCurrentToken[1],Buf,FPtr-Buf);
		   byte(FCurrentToken[0]) := FPtr-Buf;
         {$ENDIF}
		   FToken.Add(NewStr(FCurrentToken));
         FTokenPtr.Add(Buf);
		   FIndex := FToken.Count-1;
         Found := True;
      end;
		if (FPtr^<>#0) and (StrScan(FDelim,FPtr^)<>nil) then
      	Inc(FPtr);
     	FEof := FPtr^=#0;
   end;
   Result := FCurrentToken;
end;

function TRtTokenList.Prior: string;
begin
   Dec(FIndex);
	Result := Token[FIndex];
end;

function TRtTokenList.Last: string;
begin
   FIndex := Count-1;
	Result := Token[FIndex];
end;

procedure TRtTokenList.Reset;
var
	i: integer;
begin
	FPtr := FOriginalPtr;
   for i:=FToken.Count-1 downto 0 do
   	if FToken[i]<>nil then
	   	DisposeStr(PString(FToken[i]));
   FToken.Clear;
   if FOriginalPtr<>nil then
		FTLen := StrLen(FOriginalPtr);
	FDLen := StrLen(FDelim);
   FCurrentToken := '';
   FCount := 0;
   FIndex := -1;
   FEof := False;
end;

end.

